package jdk.management.resource.internal.inst;

import java.util.concurrent.ExecutionException;
import jdk.management.resource.internal.FutureWrapper;
import jdk.management.resource.internal.CompletionHandlerWrapper;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Future;
import java.nio.channels.CompletionHandler;
import java.nio.ByteBuffer;
import jdk.internal.instrumentation.InstrumentationMethod;
import jdk.management.resource.ResourceRequest;
import sun.misc.JavaIOFileDescriptorAccess;
import java.io.IOException;
import jdk.management.resource.ResourceRequestDeniedException;
import jdk.management.resource.internal.ApproverGroup;
import jdk.management.resource.internal.ResourceIdImpl;
import sun.misc.SharedSecrets;
import java.nio.channels.AsynchronousFileChannel;
import sun.nio.ch.ThreadPool;
import java.io.FileDescriptor;
import jdk.internal.instrumentation.InstrumentationTarget;

@InstrumentationTarget("sun.nio.ch.SimpleAsynchronousFileChannelImpl")
public final class SimpleAsynchronousFileChannelImplRMHooks
{
    protected final FileDescriptor fdObj;
    protected volatile boolean closed;
    
    public SimpleAsynchronousFileChannelImplRMHooks() {
        this.fdObj = null;
    }
    
    @InstrumentationMethod
    public static AsynchronousFileChannel open(FileDescriptor fileDescriptor, boolean b, boolean b2, ThreadPool threadPool) {
        AsynchronousFileChannel open = open(fileDescriptor, b, b2, threadPool);
        JavaIOFileDescriptorAccess javaIOFileDescriptorAccess = SharedSecrets.getJavaIOFileDescriptorAccess();
        long handle;
        try {
            handle = javaIOFileDescriptorAccess.getHandle(fileDescriptor);
            if (handle == -1L) {
                handle = (long)javaIOFileDescriptorAccess.get(fileDescriptor);
            }
        }
        catch (UnsupportedOperationException ex) {
            handle = (long)javaIOFileDescriptorAccess.get(fileDescriptor);
        }
        ResourceIdImpl of = ResourceIdImpl.of(Long.valueOf(handle));
        ResourceRequest approver = ApproverGroup.FILEDESCRIPTOR_OPEN_GROUP.getApprover(fileDescriptor);
        boolean n = false;
        try {
            if (approver.request(1L, of) < 1L) {
                throw new ResourceRequestDeniedException("Resource limited: too many open file descriptors");
            }
            n = true;
        }
        finally {
            if (n == 0) {
                approver.request(-1L, of);
                try {
                    open.close();
                }
                catch (IOException ex2) {}
            }
        }
        boolean n2 = false;
        ResourceRequest approver2 = ApproverGroup.FILE_OPEN_GROUP.getApprover(open);
        try {
            if (approver2.request(1L, of) < 1L) {
                try {
                    open.close();
                }
                catch (IOException ex3) {}
                throw new ResourceRequestDeniedException("Resource limited: too many open files");
            }
            n2 = true;
        }
        finally {
            if (n2 == 0) {
                approver2.request(-1L, of);
                try {
                    open.close();
                }
                catch (IOException ex4) {}
            }
        }
        return open;
    }
    
    @InstrumentationMethod
     <A> Future<Integer> implRead(ByteBuffer byteBuffer, long n, A a, CompletionHandler<Integer, ? super A> completionHandler) {
        ResourceIdImpl of = ResourceIdImpl.of(this.fdObj);
        ResourceRequest approver = ApproverGroup.FILE_READ_GROUP.getApprover(this);
        int remaining = byteBuffer.remaining();
        long max;
        try {
            max = Math.max(approver.request((long)remaining, of), 0L);
            if (max < (long)remaining) {
                throw new ResourceRequestDeniedException("Resource limited: insufficient bytes approved");
            }
        }
        catch (ResourceRequestDeniedException ex) {
            if (completionHandler != null) {
                completionHandler.failed(ex, a);
                return null;
            }
            CompletableFuture<Object> completableFuture = new CompletableFuture();
            completableFuture.completeExceptionally(ex);
            return completableFuture;
        }
        CompletionHandlerWrapper completionHandlerWrapper = null;
        if (completionHandler != null) {
            completionHandlerWrapper = new CompletionHandlerWrapper(completionHandler, of, approver, max);
        }
        Future<Integer> implRead = this.implRead(byteBuffer, n, a, completionHandlerWrapper);
        if (completionHandler == null) {
            if (implRead.isDone()) {
                int intValue = 0;
                try {
                    intValue = (int)implRead.get();
                }
                catch (InterruptedException | ExecutionException ex2) {}
                approver.request(-(max - (long)Math.max(0, intValue)), of);
            }
            else {
                implRead = new FutureWrapper(implRead, of, approver, max);
            }
        }
        return implRead;
    }
    
    @InstrumentationMethod
     <A> Future<Integer> implWrite(ByteBuffer byteBuffer, long n, A a, CompletionHandler<Integer, ? super A> completionHandler) {
        ResourceIdImpl of = ResourceIdImpl.of(this.fdObj);
        ResourceRequest approver = ApproverGroup.FILE_WRITE_GROUP.getApprover(this);
        int remaining = byteBuffer.remaining();
        long max;
        try {
            max = Math.max(approver.request((long)remaining, of), 0L);
            if (max < (long)remaining) {
                throw new ResourceRequestDeniedException("Resource limited: insufficient bytes approved");
            }
        }
        catch (ResourceRequestDeniedException ex) {
            if (completionHandler != null) {
                completionHandler.failed(ex, a);
                return null;
            }
            CompletableFuture<Object> completableFuture = new CompletableFuture();
            completableFuture.completeExceptionally(ex);
            return completableFuture;
        }
        CompletionHandlerWrapper completionHandlerWrapper = null;
        if (completionHandler != null) {
            completionHandlerWrapper = new CompletionHandlerWrapper(completionHandler, of, approver, max);
        }
        Future<Integer> implWrite = this.implWrite(byteBuffer, n, a, completionHandlerWrapper);
        if (completionHandler == null) {
            if (implWrite.isDone()) {
                int intValue = 0;
                try {
                    intValue = (int)implWrite.get();
                }
                catch (InterruptedException | ExecutionException ex2) {}
                approver.request(-(max - (long)Math.max(0, intValue)), of);
            }
            else {
                implWrite = new FutureWrapper(implWrite, of, approver, max);
            }
        }
        return implWrite;
    }
    
    @InstrumentationMethod
    public void close() throws IOException {
        synchronized (this.fdObj) {
            if (this.closed) {
                return;
            }
        }
        try {
            this.close();
            ResourceIdImpl of = ResourceIdImpl.of(Integer.valueOf(SharedSecrets.getJavaIOFileDescriptorAccess().get(this.fdObj)));
            ApproverGroup.FILEDESCRIPTOR_OPEN_GROUP.getApprover(this.fdObj).request(-1L, of);
            ApproverGroup.FILE_OPEN_GROUP.getApprover(this).request(-1L, of);
        }
        finally {
            ResourceIdImpl of2 = ResourceIdImpl.of(Integer.valueOf(SharedSecrets.getJavaIOFileDescriptorAccess().get(this.fdObj)));
            ApproverGroup.FILEDESCRIPTOR_OPEN_GROUP.getApprover(this.fdObj).request(-1L, of2);
            ApproverGroup.FILE_OPEN_GROUP.getApprover(this).request(-1L, of2);
        }
    }
}
